# -*- coding: utf-8 -*-

# Standard Library
import json
import datetime
import zlib
import base64
import socket
from dateutil.relativedelta import relativedelta

# Vendor Library
from celery import group
from celery.exceptions import MaxRetriesExceededError
from django.conf import settings
from django.db.models import Q
from django.template import loader
from django.utils.datastructures import SortedDict
from django.db.models import Q
import suds
from suds.client import Client
from suds.sax.element import Element

# Project Library
from kinder.core.async.tasks import celery_app
from kinder.core.dict.models import (
    DouType, HNE, GroupCountNorm, UnitKind, GroupTypeEnumerate, WorkType)
from kinder.core.direct.models import Direct
from kinder.core.group.models import Group, Pupil, GroupStatusEnum
from kinder.core.unit.models import StructureTypes
from kinder.core.direct.models import DRS
from kinder.core.emie_unit.models import UnitDocFiles, DopService
from gisdo.utils import UnitHelper, DateMixin
from gisdo.alchemy_session import session_scope
from gisdo.index import (
    ApplicationIndex, EnrolledIndex, CapacitiesIndex, get_queue_index)
from gisdo.index.queue import get_queue_index_n
from gisdo import constants as const
from gisdo.models import (
    ReportFormRow, ReportForm, ScheduleSettings, InstallationType)
from gisdo import settings as gs_settings
from gisdo.exceptions import (
    SendReportException, DocumentCreateException,
    XMLRenderException, SOAPClientCreateException,
    NoDataInMo, ConnectingAtSendingReportException)
from gisdo.logger import Logger
from .index.direct import DirectIndex
from .utils import AgeDeltas
from .queue.queue_cache import get_queue_cache


class ReportCacheMixin(object):
    APPLICATIONS = 'applications'
    QUEUES = 'queues'
    ENROLLED = 'enrolled'
    CAPACITIES = 'capacities'

    DATA_TYPE = [
        APPLICATIONS,
        QUEUES,
        ENROLLED,
        CAPACITIES
    ]

    def __init__(self):

        self._cache = {
            self.APPLICATIONS: {},
            self.QUEUES: {},
            self.ENROLLED: {},
            self.CAPACITIES: {}
        }

    def _update_cache(self, data, data_type):
        """
        Обновление кэша
        """

        for key, value in data.iteritems():
            if not key in self._cache[data_type]:
                self._cache[data_type][key] = value
            elif isinstance(value, dict):
                for age_cat in data[key]:
                    self._cache[
                        data_type][key][age_cat] += data[key][age_cat]


class ReportService(ReportCacheMixin):
    def __init__(
            self, main_unit,
            set_progress_func=None,
            user=None):

        super(ReportService, self).__init__()

        self._unit_helper = UnitHelper(main_unit)
        self._report_form = ReportForm(
            user=user, unit=main_unit,
            in_progress=True)
        self._set_progress = set_progress_func or (lambda x: x)
        self._schedule_settings = ScheduleSettings.get_settings()

    def collect(self, send=True):
        self._report_form.save()

        # Счетчик кол-ва ошибок при отправке
        send_error_count = 0

        # На этом уровне перебираем учреждения типа МО
        current_mo_num = 1
        mo_count = len(list(self._unit_helper.get_report_units()))

        # Создаем группу параллельно выполняющихся подзадач
        # Каждая задача отвечает за подсчет показателей отдельного МО
        collect_data_tasks = group(
            self._collect_data.s(
                mo, self._unit_helper._unit, self._report_form
            ) for mo in self._unit_helper.get_report_units())

        collect_data_tasks_results = collect_data_tasks(
            queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

        mo_for_send = []

        # Обрабатываем результаты завершившихся задач
        for (mo, task_result, exception) in collect_data_tasks_results.iterate():
            self._set_progress(
                u'Обработано МО: %s из %s' % (
                    current_mo_num, mo_count))

            now = str(datetime.datetime.now())

            try:
                # Выбрасываем отложенное исключение
                if exception:
                    raise exception
            except Exception as e:
                self._set_progress(
                    values={
                        now: (
                            u'Во время сбора данных по'
                            u' %s произошла ошибка: %s' % (mo.name, e.message))
                    })
                current_mo_num += 1
                continue
            else:
                self._update_data(task_result)

            self._set_progress(
                values={
                    now: (
                        u'Данные по %s успешно собраны' % mo.name)
                })

            current_mo_num += 1

            if send:
                mo_for_send.append(mo)

        current_mo_num = 1

        if send:
            self._set_progress(u'Подготовка к отправке данные МО')

            minutes, max_retries = (
                self._schedule_settings.get_interval_and_count_retries())

            # Создаем группу параллельно выполняющихся подзадач
            # Каждая задача отвечает за отправку показателей отдельного МО
            # Замечание: если подзадача выполнилась с ошибкой связанной с
            #   недоступностью сервера, то она будет перезапущена через
            #   "countdown" секунд максимум "max_retries" раз.
            #   И только по завершению всех возможных попыток всех подзадач
            #   группа подзадач считается выполненной.
            send_data_tasks = group(
                self.send_mo_data.s(
                    mo, self._unit_helper._unit,
                    self._report_form, self._schedule_settings,
                    countdown=minutes * 60, max_retries=max_retries
                ) for mo in mo_for_send)

            # добавляем задержку каждой подзадачи на 1 секунду
            send_data_tasks_results = send_data_tasks.skew(
                start=1, stop=len(mo_for_send))(
                queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

            for (mo, exception) in send_data_tasks_results.iterate():
                self._set_progress(
                    u'Отправлены данные МО: %s из %s' % (
                        current_mo_num, mo_count))

                now = str(datetime.datetime.now())

                try:
                    if exception:
                        raise exception
                except Exception as e:
                    self._set_progress(
                        values={
                            now: (
                                u'При передаче данных по %s '
                                u'произошла ошибка: %s' % (mo.name, e.message))
                        })
                    current_mo_num += 1
                    send_error_count += 1
                    continue

                self._set_progress(
                    values={
                        now: (
                            u'Данные по %s успешно отправлены' % mo.name)
                    })

                current_mo_num += 1

        if self._unit_helper.is_region:
            ReportFormRow.objects.get_or_create(
                report=self._report_form,
                unit=self._unit_helper.get_main_unit(),
                applications=json.dumps(self._cache[self.APPLICATIONS]),
                queues=json.dumps(self._cache[self.QUEUES]),
                enrolled=json.dumps(self._cache[self.ENROLLED]),
                capacities=json.dumps(self._cache[self.CAPACITIES]))

        self._report_form.in_progress = False

        # Если все пакеты были отправлены то считаем что отчет отправлен
        if send and send_error_count == 0:
            self._report_form.sent_time = datetime.datetime.now()

        self._report_form.date = datetime.datetime.now()
        self._report_form.save()

    @celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
    def _collect_data(mo, main_report_unit, report_form):
        unit_helper = UnitHelper(main_report_unit)

        data = None
        exception = None

        try:
            data = DataCollector(
                mo, unit_helper).collect(report_form)
        # Отложенное исключение
        except Exception as e:
            exception = e
        return (mo, data, exception)

    def _update_data(self, mo_data):
        if self._unit_helper.is_region:
            self._update_cache(mo_data[self.APPLICATIONS], self.APPLICATIONS)
            self._update_cache(mo_data[self.QUEUES], self.QUEUES)
            self._update_cache(mo_data[self.ENROLLED], self.ENROLLED)
            self._update_cache(mo_data[self.CAPACITIES], self.CAPACITIES)

    @celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
    def send_mo_data(mo, main_report_unit, report_form, settings, **kw):
        unit_helper = UnitHelper(main_report_unit)

        exception = None
        try:
            Transmitter(
                report_form,
                settings,
                unit_helper).send(mo)
        except ConnectingAtSendingReportException as exc:
            # если произошла ошибка связанная с соединением
            # сервера сервиса, то перезапускаем таску через
            # "countdown" максимум "max_retries" раз
            try:
                ReportService.send_mo_data.retry(
                    exc=exc,
                    max_retries=kw.get('max_retries'),
                    countdown=kw.get('countdown'))
            except MaxRetriesExceededError as e:
                # если все возможные попытки не увенчались успехом,
                # то сообщаем возвращаем исходную ошибку "exc"
                exception = exc
        except Exception as e:
            exception = e

        return (mo, exception)


class Transmitter(object):

    def __init__(self, report_form, schedule_settings, unit_helper):

        self._report_form = report_form
        self._schedule_settings = schedule_settings
        self._unit_helper = unit_helper

    def send(self, unit=None):

        if not unit:
            unit = self._unit_helper.get_main_unit()
        elif unit.kind.id not in [UnitKind.MO, UnitKind.REGION]:
            raise SendReportException()

        # Список МО по которым не удалось отправить данные
        not_send_mo = []
        for mo in unit.get_descendants(
                include_self=True).filter(kind__id=UnitKind.MO):

            try:
                document = DocumentService(
                    mo, self._report_form.date,
                    self._schedule_settings,
                    self._unit_helper,
                    self._report_form).create()
            except NoDataInMo as e:
                not_send_mo.append(mo.name)
                Logger.add_record(e.message, level=Logger.WARNING)
                continue
            except Exception as e:
                Logger.add_record(e.message, level=Logger.ERROR)
                raise DocumentCreateException()

            try:
                xml_content = TemplateLoader(
                    self._schedule_settings).load(document)
            except Exception as e:
                Logger.add_record(e.message, level=Logger.ERROR)
                raise XMLRenderException()

            BroadCaster(self._schedule_settings).send(
                xml_content)

        return not_send_mo


class DataCollector(ReportCacheMixin):
    def __init__(self, mo, unit_helper):

        super(DataCollector, self).__init__()

        self._mo = mo
        self._unit_helper = unit_helper

    def collect(self, report_form):

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        # Собираем по ДОУ
        for dou in dou_list:
            self._dou_collect(report_form, dou)

        # Данные по МО лежат в кэше
        ReportFormRow.objects.get_or_create(
            applications=json.dumps(self._cache[self.APPLICATIONS]),
            queues=json.dumps(self._cache[self.QUEUES]),
            enrolled=json.dumps(self._cache[self.ENROLLED]),
            capacities=json.dumps(self._cache[self.CAPACITIES]),
            report=report_form,
            unit=self._mo)

        return self._cache

    def _dou_collect(self, report_form, dou):
        import time
        t = time.time()
        applications = self._collect_applications_data(dou)
        # print "1", time.time() - t
        # t = time.time()
        queues = self._collect_queue_data(dou)
        # print "2", time.time() - t
        # t = time.time()
        enrolled = self._collect_enrolled_data(dou, queues)
        # print "3", time.time() - t
        # t = time.time()
        capacities = self._collect_capacities(dou)
        # print "4", time.time() - t
        print time.time() - t

        ReportFormRow.objects.get_or_create(
            report=report_form, unit=dou,
            applications=json.dumps(applications),
            queues=json.dumps(queues),
            enrolled=json.dumps(enrolled),
            capacities=json.dumps(capacities))

    def _collect_applications_data(self, dou):

        # Словарь с id-шниками показателей
        applications = SortedDict([
            ('1', SortedDict()),
            ('1.1', SortedDict()),
            ('2', SortedDict()),
            ('3', SortedDict()),
            ('4', SortedDict()),
            ('4.1', SortedDict()),
            ('4.2', SortedDict()),
            ('5', SortedDict()),
            ('6', SortedDict()),
        ])

        index_4 = get_queue_index_n("4", dou)
        index_4_1 = get_queue_index_n("4.1", dou)
        index_4_2 = get_queue_index_n("4.2", dou)
        index_5 = get_queue_index_n("5", dou)

        with session_scope() as session:
            index = ApplicationIndex(dou, session)

            for value in const.AGE_CATEGORIES_FULL.values():

                applications['2'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_DELIVERY, portal=1,
                    age_range=value, on_date=datetime.date.today)

                applications['3'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                    age_range=value, on_date=datetime.date.today)

                applications['1.1'][value] = (
                    applications['2'][value] +
                    applications['3'][value])

                on_date_filter = get_age_filter(
                    value, datetime.date.today)

                applications["4"][value] = len([
                    declaration
                    for declaration in index_4
                    if on_date_filter(declaration)])

                applications["4.1"][value] = len([
                    declaration
                    for declaration in index_4_1
                    if on_date_filter(declaration)])

                applications["4.2"][value] = len([
                    declaration
                    for declaration in index_4_2
                    if on_date_filter(declaration)])

                applications["5"][value] = len([
                    declaration
                    for declaration in index_5
                    if on_date_filter(declaration)])

                # Общее количество заявлений на постановку в очередь
                # в группы кратковременного пребывания
                applications['6'][value] = index.get_count(
                    index_type=ApplicationIndex.COUNT_BY_GROUP_TYPE,
                    age_range=value, on_date=datetime.date.today)

            applications['1'][const.ALL] = get_queue_index('1', dou)

            self._update_cache(applications, self.APPLICATIONS)

            return applications

    def _collect_queue_data(self, dou):

        queues = SortedDict([
            ('7', SortedDict()), ('7.1', SortedDict()),
            ('7.2', SortedDict()), ('7.3', SortedDict()),
            ('7.4', SortedDict()), ('8', SortedDict()),
            ('9', SortedDict()), ('9.1', SortedDict()),
            ('10', SortedDict()), ('11', SortedDict()),
            ('12', SortedDict()), ('12.1', SortedDict()),
            ('13', SortedDict()),
            ('13.1', SortedDict()), ('14', SortedDict()),
            ('15', SortedDict()), ('16', SortedDict()),
            ('17', SortedDict()), ('17.1', SortedDict()),
            ('18', SortedDict()), ('18.1', SortedDict()),
            ('18.2', SortedDict()), ('18.3', SortedDict())
        ])

        index_7 = get_queue_index_n("7", dou)
        index_7_1 = get_queue_index_n("7.1", dou)
        index_7_2 = get_queue_index_n("7.2", dou)
        index_7_3 = get_queue_index_n("7.3", dou)
        index_7_4 = get_queue_index_n('7.4', dou)
        index_18_1 = get_queue_index_n('18.1', dou)
        index_18_3 = get_queue_index_n('18.3', dou)
        index_20 = get_queue_index_n("20", dou)
        index_9 = get_queue_index_n("9", dou)
        index_12 = get_queue_index_n("12", dou)
        index_12_1 = get_queue_index_n("12.1", dou)
        index_13 = get_queue_index_n("13", dou)
        index_13_1 = get_queue_index_n("13.1", dou)
        index_14 = get_queue_index_n("14", dou)
        index_15 = get_queue_index_n("15", dou)
        index_16 = get_queue_index_n("16", dou)
        index_17 = get_queue_index_n("17", dou)
        index_17_1 = get_queue_index_n("17.1", dou)
        index_18 = get_queue_index_n("18", dou)
        index_18_2 = get_queue_index_n("18.2", dou)
        index_10 = get_queue_index_n("10", dou)
        index_11 = get_queue_index_n("11", dou)

        queues['10'][const.ALL] = 0
        queues['11'][const.ALL] = 0
        for value in const.AGE_CATEGORIES_FULL.values():

            current_year_filter = get_age_filter(
                value, DateMixin._get_current_calendar_year)
            on_date_filter = get_age_filter(
                value, datetime.date.today)
            current_year_first_day = get_age_filter(
                value,
                datetime.date(datetime.date.today().year, 1, 1))

            # Отложенный спрос
            queues["7"][value] = len([
                declaration
                for declaration in index_7
                if current_year_filter(declaration)])

            queues["7.1"][value] = len([
                declaration
                for declaration in index_7_1
                if on_date_filter(declaration)])

            queues["7.2"][value] = len([
                declaration
                for declaration in index_7_2
                if on_date_filter(declaration)])

            queues["7.3"][value] = len([
                declaration
                for declaration in index_7_3
                if on_date_filter(declaration)])

            queues["7.4"][value] = len([
                declaration
                for declaration in index_7_4
                if on_date_filter(declaration)])

            # Актуальный спрос
            queues["18.1"][value] = len([
                declaration
                for declaration in index_18_1
                if on_date_filter(declaration)])

            queues["18.3"][value] = len([
                declaration
                for declaration in index_18_3
                if on_date_filter(declaration)])

            # Актуальный спрос + Желает сменить ДОО
            indx20 = len([
                declaration
                for declaration in index_20
                if on_date_filter(declaration)])
            queues['8'][value] = (
                queues['18.1'][value] + queues['18.3'][value] +
                indx20)

            # Количество детей в очереди, имеющих право
            # на внеочередное и первоочередное зачисление
            queues["9"][value] = len([
                declaration
                for declaration in index_9
                if current_year_filter(declaration)])

            # Количество детей в очереди с ограниченными
            # возможностями по здоровью
            queues["12"][value] = len([
                declaration
                for declaration in index_12
                if on_date_filter(declaration)])

            queues["12.1"][value] = len([
                declaration
                for declaration in index_12_1
                if on_date_filter(declaration)])

            queues["13"][value] = len([
                declaration
                for declaration in index_13
                if on_date_filter(declaration)])

            queues["13.1"][value] = len([
                declaration
                for declaration in index_13_1
                if on_date_filter(declaration)])

            queues["14"][value] = len([
                declaration
                for declaration in index_14
                if on_date_filter(declaration)])

            queues["15"][value] = len([
                declaration
                for declaration in index_15
                if on_date_filter(declaration)])

            queues["16"][value] = len([
                declaration
                for declaration in index_16
                if on_date_filter(declaration)])

            queues["17"][value] = len([
                declaration
                for declaration in index_17
                if on_date_filter(declaration)])

            queues["17.1"][value] = len([
                declaration
                for declaration in index_17_1
                if on_date_filter(declaration)])

            # Актуальный спрос
            queues["18"][value] = len([
                declaration
                for declaration in index_18
                if current_year_filter(declaration)])

            queues["18.2"][value] = len([
                declaration
                for declaration in index_18_2
                if current_year_first_day(declaration)])

            # Количество детей в очереди имеющих право зачисления на основе
            # федеральных и региональных льгот.
            current_10_value = len([
                declaration
                for declaration in index_10
                if on_date_filter(declaration)])
            queues['10'][const.ALL] += current_10_value
            current_11_value = len([
                declaration
                for declaration in index_11
                if on_date_filter(declaration)])

            queues['11'][const.ALL] += current_11_value
            queues['9.1'][value] = (
                current_10_value + current_11_value)

        self._update_cache(queues, self.QUEUES)

        return queues

    def _collect_enrolled_data(self, dou, queues):

        enrolled = SortedDict([
            ('19', SortedDict()), ('19.1', SortedDict()),
            ('19.2', SortedDict()), ('19.3', SortedDict()),
            ('20', SortedDict()), ('20.1', SortedDict()),
            ('20.2', SortedDict()), ('20.3', SortedDict()),
            ('20.4', SortedDict()), ('21', SortedDict()),
            ('22', SortedDict()), ('23', SortedDict()),
            ('24', SortedDict()), ('25', SortedDict()),
            ('26', SortedDict()), ('27', SortedDict()),
            ('28', SortedDict()), ('29', SortedDict()),
            ('30', SortedDict()), ('30.1', SortedDict()),
            ('30.2', SortedDict()),
        ])

        with session_scope() as session:
            index = EnrolledIndex(dou, session)
            direct_index = DirectIndex(dou)

            index_19 = index.get_count()
            index_19_1 = index.get_count()
            index_19_2 = index.get_count()
            index_20 = get_queue_index_n("20", dou)
            index_20_1 = get_queue_index_n("20.1", dou)
            index_20_2 = get_queue_index_n("20.2", dou)
            index_20_3 = get_queue_index_n("20.3", dou)
            index_22 = index.get_count(
                index_type=EnrolledIndex.COUNT_DISABLED)
            index_23 = index.get_count(
                index_type=EnrolledIndex.COUNT_CORRECT)
            index_24 = index.get_count(
                index_type=EnrolledIndex.COUNT_HEALTH)
            index_25 = index.get_count(
                index_type=EnrolledIndex.COUNT_BY_CARE)
            index_26 = index.get_count(
                index_type=EnrolledIndex.COUNT_FULL_DAY_STAY)
            index_27 = index.get_count(
                index_type=EnrolledIndex.COUNT_SHORT_STAY)
            index_30 = get_queue_index_n("30", dou)
            index_30_2 = get_queue_index_n("30.2", dou)

            for value in const.AGE_CATEGORIES_FULL.values():
                current_year_filter = get_age_filter(
                    value, DateMixin._get_current_calendar_year)
                current_year_enrolled_filter = get_enrolled_age_filter(
                    value, DateMixin._get_current_calendar_year)
                on_date_enrolled_filter = get_enrolled_age_filter(
                    value, datetime.date.today)
                next_calendar_year_enrolled = get_enrolled_age_filter(
                    value, DateMixin.get_next_calendar_year)

                # Общая численность детей зачисленных в ДОО
                enrolled['19'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_19
                    if current_year_enrolled_filter(date_of_birth)
                ])
                # enrolled['19'][value] = index.get_count(
                #     age_range=value,
                #     on_date=DateMixin._get_current_calendar_year)
                enrolled['19.1'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_19_1
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['19.1'][value] = index.get_count(
                #     age_range=value, on_date=datetime.date.today)

                enrolled['19.2'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_19_2
                    if next_calendar_year_enrolled(date_of_birth)
                ])
                # enrolled['19.2'][value] = index.get_count(
                #     age_range=value,
                #     on_date=DateMixin.get_next_calendar_year)

                enrolled['19.3'][value] = direct_index.get_count(
                    DirectIndex.FILTER_193, age_range=value,
                    on_date=datetime.date.today)

                on_date_filter = get_age_filter(
                    value, datetime.date.today)
                enrolled["20"][value] = len([
                    declaration
                    for declaration in index_20
                    if on_date_filter(declaration)])

                enrolled["20.1"][value] = len([
                    declaration
                    for declaration in index_20_1
                    if on_date_filter(declaration)])

                enrolled["20.2"][value] = len([
                    declaration
                    for declaration in index_20_2
                    if on_date_filter(declaration)])

                enrolled["20.3"][value] = len([
                    declaration
                    for declaration in index_20_3
                    if on_date_filter(declaration)])

                # Общая численность детей, зачисленных в группы для
                # детей с ограниченными возможностями здоровью
                enrolled['22'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_22
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['22'][value] = index.get_count(
                #     index_type=EnrolledIndex.COUNT_DISABLED,
                #     age_range=value, on_date=datetime.date.today)

                # Общее количество детей зачисленных в
                # оздоровительные и компенсирующие группы
                enrolled['23'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_23
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['23'][value] = index.get_count(
                #     index_type=EnrolledIndex.COUNT_CORRECT,
                #     age_range=value, on_date=datetime.date.today)

                enrolled['24'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_24
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['24'][value] = index.get_count(
                #     index_type=EnrolledIndex.COUNT_HEALTH,
                #     age_range=value, on_date=datetime.date.today)

                enrolled['25'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_25
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['25'][value] = index.get_count(
                #     index_type=EnrolledIndex.COUNT_BY_CARE,
                #     age_range=value, on_date=datetime.date.today)

                # Общая численность детей, зачисленных в
                # группы круглосуточного пребывания детей
                enrolled['26'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_26
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['26'][value] = index.get_count(
                #     index_type=EnrolledIndex.COUNT_FULL_DAY_STAY,
                #     age_range=value, on_date=datetime.date.today)

                # Группы кратковременного пребывания
                enrolled['27'][value] = len([
                    1
                    for _, _, _, date_of_birth in index_27
                    if on_date_enrolled_filter(date_of_birth)
                ])
                # enrolled['27'][value] = index.get_count(
                #     age_range=value,
                #     index_type=EnrolledIndex.COUNT_SHORT_STAY,
                #     on_date=datetime.date.today)
                enrolled['28'][value] = enrolled['27'][value]

                i30 = len([
                    declaration
                    for declaration in index_30
                    if current_year_filter(declaration)])

                enrolled['30'][value] = (
                    enrolled["19"][value] +
                    queues["18"][value] +
                    i30
                )

                enrolled['30.1'][value] = (
                    queues['18.1'][value] + enrolled['19.1'][value] +
                    enrolled['19.3'][value]
                )

                next_year_filter = get_age_filter(
                    value, DateMixin._get_next_calendar_year)
                i30_2 = len([
                    declaration
                    for declaration in index_30_2
                    if next_year_filter(declaration)])
                enrolled['30.2'][value] = (
                    enrolled['19.2'][value] +
                    queues['18.2'][value] +
                    i30_2
                )

            enrolled['20.4'][const.ALL] = get_queue_index("20.4", dou)
            # Общее количество детей зачисленных в семейные,
            # оздоровительные и компенсирующие группы
            enrolled['21'][const.ALL] = index.get_count(
                index_type=EnrolledIndex.COUNT_FAMILY,
                on_date=datetime.date.today).count()
            enrolled['29'][const.ALL] = index.get_count(
                index_type=EnrolledIndex.COUNT_DECREASE_CONTINGENT,
                on_date=datetime.date.today).count()

        self._update_cache(enrolled, self.ENROLLED)

        return enrolled

    def _collect_capacities(self, dou):

        capacities = SortedDict([
            ('31', SortedDict()), ('31.1', SortedDict()),
            ('31.2', SortedDict()), ('31.3', SortedDict()),
            ('31.4', SortedDict()), ('32', SortedDict()),
            ('33', SortedDict())
        ])

        index = CapacitiesIndex(dou)
        direct_index = DirectIndex(dou)

        for value in const.AGE_CATEGORIES_EIGHT.values():
            capacities['31'][value] = index.get_count(
                age_category=value)
            capacities['31.1'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.HEALTH_NEED)
            capacities['31.2'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.COMPENSATING)
            capacities['31.3'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.HEALTH)
            capacities['31.4'][value] = index.get_count(
                age_category=value,
                index_type=CapacitiesIndex.SHORT_STAY)

        capacities['32'][const.ALL] = direct_index.get_count(
            DirectIndex.FILTER_32,
            on_date=datetime.date.today)
        capacities['33'][const.ALL] = get_queue_index("33", dou)

        self._update_cache(capacities, self.CAPACITIES)
        return capacities


class DocumentService(object):
    def __init__(
            self, mo, date, schedule_settings,
            unit_helper, report_form):

        self._mo = mo
        self._date = date.date()
        self._gisdo_settings = schedule_settings
        self._unit_helper = unit_helper
        self._report_form = report_form
        self._document = {}

    def _create_header(self):
        """
        Формирование данных для заголовка
        """

        self._document.update(
            report_date=self._date.isoformat(),
            send_time=datetime.datetime.now().isoformat(),
            system_name=self._gisdo_settings.system_name,
            version=settings.KINDER_VERSION,
            email=self._gisdo_settings.email,
            install_type=InstallationType.values.get(
                self._gisdo_settings.installation_type,
                InstallationType.values[InstallationType.REGIONAL]),
            vendor=u'ЗАО БАРС Груп')

    def _get_municipality_info(self):

        result = {
            'OKTMO': self._mo.octmo or ''
        }

        if self._mo.epgu_link:
            result.update({'epguLink': self._mo.epgu_link})
        else:
            result.update({'epguLink': u"http://www.gosuslugi.ru/"})

        if self._mo.rpgu_link:
            result.update({'rpguLink': self._mo.rpgu_link})
        else:
            result.update({'rpguLink': u"нет"})

        today = datetime.date.today()
        result["family_0_3"] = Pupil.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=3),
            children__date_of_birth__lt=today,
            temporary_deduct=False,
            grup__unit__lft__gte=self._mo.lft,
            grup__unit__rght__lte=self._mo.rght,
            grup__is_family=True).count()
        result["family_3_7"] = Pupil.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=7),
            children__date_of_birth__lte=today - relativedelta(years=3),
            temporary_deduct=False,
            grup__unit__lft__gte=self._mo.lft,
            grup__unit__rght__lte=self._mo.rght,
            grup__is_family=True).count()

        return result

    @staticmethod
    def _get_unit_info(dou):

        return OrganizationInfo.get_info(dou)

    @staticmethod
    def _get_contact_info(dou):

        return ContactInfo.get_info(dou)

    @staticmethod
    def _get_groups(dou):

        return GroupsInfo.get_info(dou)

    @staticmethod
    def _get_additional_info(dou, model=UnitDocFiles):

        extensions = model.objects.filter(
            unit=dou).values_list('name', flat=True)

        if not extensions.exists():
            return [u'Нет']
        else:
            return extensions

    @staticmethod
    def _transform(counter):

        return CounterProxy.transform(counter)

    def _get_counters(self, dou):

        rows = ReportFormRow.objects.filter(
            unit=dou, report=self._report_form)

        counters = []
        for row in rows:
            counter = {}
            counter.update(json.loads(row.applications))
            counter.update(json.loads(row.queues))
            counter.update(json.loads(row.enrolled))
            counter.update(json.loads(row.capacities))
            counters.extend(DocumentService._transform(counter))

        return counters

    @staticmethod
    def _get_group_counters(dou):

        return Group.objects.filter(unit=dou, status=GroupStatusEnum.FACT).values('id')

    def _create_reports(self):

        reports = []

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        for dou in dou_list:
            report = {}
            report.update(self._get_municipality_info())
            report.update({
                'unit': DocumentService._get_unit_info(dou),
                'contact': DocumentService._get_contact_info(dou),
                'groups': DocumentService._get_groups(dou),
                'additionalEducations': DocumentService._get_additional_info(
                    dou),
                'features': DocumentService._get_additional_info(
                    dou, model=DopService),
                'counters': self._get_counters(dou),
                'group_counters': DocumentService._get_group_counters(dou)
            })

            reports.append(report)

        self._document.update({'reports': reports})

    def create(self):

        self._create_header()
        self._create_reports()

        return self._document


class CounterProxy(object):
    @staticmethod
    def transform(counter):

        result = []
        for index_id in counter:

            index = {
                'id': index_id,
                'indexes': []
            }

            for age_cat in counter[index_id]:
                index['indexes'].append({
                    'category': age_cat,
                    'count': counter[index_id][age_cat]
                })

            result.append(index)

        return result


class OVZInfoMixin(object):
    NOT_OVZ_GROUPS = u'Нет групп ОВЗ'
    DIFFERENT_OVZ_GROUPS = u'Комбинированные группы ОВЗ'
    OVZ_TYPES = {
        HNE.DEAFNESS: u'Нарушения слуха',
        HNE.HARDOFHEARTING: u'Нарушения слуха',
        HNE.SPEACH: u'Нарушения речи',
        HNE.PHONETICS: u'Нарушения речи',
        HNE.AMAUROSIS: u'Нарушения зрения',
        HNE.BLINDNESS: u'Нарушения зрения',
        HNE.BACKLIGHT: u'Умственная отсталость',
        HNE.BACKHARD: u'Умственная отсталость',
        HNE.BACK: u'Задержка психического развития',
        HNE.DISABLEMENT: u'Нарушения опорно-двигательного аппарата',
        HNE.INVALIDITY: u'Сложные дефекты',
        HNE.OTHER: u'Иные нарушения в развитии'
    }


class OrganizationInfo(OVZInfoMixin):
    ORGANIZATION_TYPES = {
        DouType.GOVERNMENT: u'Государственная',
        DouType.MUNICIPAL: u'Муниципальная',
        DouType.NOT_GOVERNMENT_WITH_LICENSE: u'Частная (есть лицензия)',
        DouType.NOT_GOVERNMENT_WITHOUT_LICENSE: u'Частная (без лицензии)',
        DouType.IP_WITH_LICENSE: (
            u'Индивидуальный предприниматель, ИП (есть лицензия)'),
        DouType.IP_WITHOUT_LICENSE: (
            u'Индивидуальный предприниматель, ИП (без лицензии)'),
        DouType.DEPARTMENTAL: u'Ведомственная',
        DouType.OTHER: u'Другие типы'
    }

    BUILD_IN_PROGRESS = 1
    CLOSED = 2
    IN_CAPITAL_REPAIR = 3
    WORK = 4

    STATUSES = {
        BUILD_IN_PROGRESS: u'Ожидает открытия',
        CLOSED: u'Деятельность приостановлена',
        IN_CAPITAL_REPAIR: u'Капитальный ремонт',
        WORK: u'Функционирует'
    }

    @classmethod
    def _get_dou_type(cls, dou):

        dou_type = dou.dou_type

        if not dou_type:
            code = DouType.GOVERNMENT
        else:
            code = dou_type.code

        return cls.ORGANIZATION_TYPES[code]

    @classmethod
    def _get_dou_status(cls, dou):

        status = cls.STATUSES[cls.WORK]

        if dou.closed:
            status = cls.STATUSES[cls.CLOSED]
        elif dou.build_in_progress:
            status = cls.STATUSES[cls.BUILD_IN_PROGRESS]
        elif dou.in_capital_repair:
            status = cls.STATUSES[cls.IN_CAPITAL_REPAIR]

        return status

    @classmethod
    def _get_dou_structure_type(cls, dou):
        structure_type = dou.structure_type
        if structure_type == StructureTypes.OTHER:
            structure_type = StructureTypes.ELEMENTARY

        return structure_type

    @classmethod
    def get_info(cls, dou):

        result = {
            'code': dou.gisdo.doo_identity,
            'name': dou.name,
            'type': cls._get_dou_type(dou),
            'status': cls._get_dou_status(dou),
            'structure_type': cls._get_dou_structure_type(dou),
            'director': dou.boss_fio or u"нет",
        }

        if dou.boss_fio:
            result.update({'boss_fio': dou.boss_fio})

        return result


class ContactInfo(object):
    @classmethod
    def get_info(cls, dou):

        result = {}

        if dou.address_house_guid:
            result.update({'fiasHouseGuid': dou.address_house_guid})
        elif dou.address_street:
            result.update({'fiasHouseGuid': dou.address_street})
        elif dou.address_place:
            result.update({'fiasHouseGuid': dou.address_place})

        result.update(
            {"plainAddress": dou.address_full
                if dou.address_full else u"нет"})
        result.update({"website": dou.site if dou.site else u"нет"})
        result.update({"email": dou.email if dou.email else u"нет"})
        result.update({"phone": dou.telephone if dou.telephone else u"нет"})

        return result


class GroupsInfo(OVZInfoMixin):

    OVZ_VALUES = {
        HNE.DEAFNESS: 1,
        HNE.HARDOFHEARTING: 1,
        HNE.SPEACH: 2,
        HNE.PHONETICS: 2,
        HNE.BLINDNESS: 3,
        HNE.BACKLIGHT: 4,
        HNE.BACKHARD: 4,
        HNE.BACK: 5,
        HNE.DISABLEMENT: 6,
        HNE.INVALIDITY: 7,
        'other': 8
    }

    WELLNESS = {
        HNE.PHTHISIS: 1,
        HNE.SICK: 2,
        'other': 3
    }

    WORKTYPE = {
        WorkType.SHORT: 1,
        WorkType.ABREV: 2,
        WorkType.FULL: 3,
        WorkType.EXTEND: 4,
        WorkType.ALLDAY: 5,
    }

    @classmethod
    def _get_capacity(cls, group):

        if group.type and group.type.code == GroupTypeEnumerate.DEV:
            return 0

        return Group.objects.extra(select={
            'count_norm': """
                SELECT "{t1}"."count" FROM "{t1}"
                  WHERE (CASE WHEN "{t2}"."health_need_id" IS NOT NULL
                      THEN "{t1}"."health_need_id"="{t2}"."health_need_id"
                      AND "{t1}"."age_cat_id"="{t2}"."age_cat_id"
                      ELSE "{t1}"."age_cat_id"="{t2}"."age_cat_id"
                      and "{t1}"."cathegory_id"="{t2}"."cathegory_id"
                      END) limit 1
            """.format(
                t1=GroupCountNorm._meta.db_table,
                t2=Group._meta.db_table,)
        })[0].count_norm or 0

    @classmethod
    def _get_enrolled(cls, group):

        return Pupil.objects.filter(
            grup=group, temporary=False,
            temporary_deduct=False).count()

    @classmethod
    def _get_group_ovz_type(cls, group):
        if group.health_need:
            return cls.OVZ_VALUES.get(
                group.health_need.code, cls.OVZ_VALUES['other'])
        else:
            return cls.OVZ_VALUES['other']

    @classmethod
    def _get_group_wellness(cls, group):
        if group.health_need:
            return cls.WELLNESS.get(
                group.health_need.code, cls.WELLNESS['other'])
        else:
            return cls.WELLNESS['other']

    @classmethod
    def _get_group_orientation(cls, group, age_from, age_to):
        """Возвращает значение для тэга orientation(Направленность группы)"""
        orientation_value = {
            GroupTypeEnumerate.DEV: {'orientation': 1},
            GroupTypeEnumerate.COMP: {
                'orientation': 2,
                'ovzType': cls._get_group_ovz_type(group)
            },
            GroupTypeEnumerate.COMBI:  {
                'orientation': 3,
                'ovzType': cls._get_group_ovz_type(group)
            },
            GroupTypeEnumerate.HEALTH: {
                'orientation': 4,
                'wellness': cls._get_group_wellness(group)
            },
            'sub_age': {"orientation": 5},
            'is_family': {"orientation": 7},
            'is_care': {"orientation": 6},
        }

        if (group.type and group.type.code == GroupTypeEnumerate.COMP
                or group.type.code == GroupTypeEnumerate.COMBI
                or group.type.code == GroupTypeEnumerate.HEALTH):
            return orientation_value[group.type.code]

        if group.is_family:
            return orientation_value['is_family']

        if group.is_care:
            return orientation_value['is_care']
        # верхняя граница подкатегории группы
        if age_to <= 3:
            return orientation_value['sub_age']

        return orientation_value[GroupTypeEnumerate.DEV]

    @classmethod
    def _get_group_work_time(cls, group):
        """Возвращает значение режима работы группы для фед. отчет."""
        def normalize_work_type_code(code):
            if code in [WorkType.ABREV, WorkType.ALLDAY,
                        WorkType.EXTEND, WorkType.FULL, WorkType.SHORT]:
                return code
            else:
                try:
                    torn_code = code.split("-")
                except AttributeError:
                    # Код может быть не заполнен
                    return WorkType.FULL

            map_code = torn_code[0]
            map_code = map_code.replace(",", ".")

            try:
                numeric_map_code = float(map_code)
            except ValueError:
                # Если код не число, то передаем "Полного дня (10,5-12 часов)"
                return WorkType.FULL

            if 1 <= numeric_map_code < 8:
                code_compliance = WorkType.SHORT
            elif 8 <= numeric_map_code < 10.5:
                code_compliance = WorkType.ABREV
            elif 10.5 <= numeric_map_code < 12.5:
                code_compliance = WorkType.FULL
            elif 12.5 <= numeric_map_code < 24:
                code_compliance = WorkType.EXTEND
            elif numeric_map_code == 24:
                code_compliance = WorkType.ALLDAY
            else:
                code_compliance = WorkType.FULL

            return code_compliance

        if group.work_type:
            normalize_code = normalize_work_type_code(group.work_type.code)
            return cls.WORKTYPE[normalize_code]
        else:
            return cls.WORKTYPE[WorkType.ALLDAY]

    @classmethod
    def _get_group_acitivity(cls, group):
        """Возвращает значения вида деятельности для фед. отчет.

         1 - реализуются образовательные программы
         дошкольного образования (в остальных случаях)
         2 - осуществляется только деятельность по присмотру и уходу."""
        return 2 if group.is_care else 1

    @classmethod
    def _get_count_ovz_child(cls, group):
        """Возвращает количество детей с ОВЗ в группе.

         Если группа компенсирующая, оздоровительная
         берем количество детей в группе
         иначе считаем количество детей имеется ОВЗ."""
        if (group.type and group.type.code == GroupTypeEnumerate.COMP
                or group.type.code == GroupTypeEnumerate.HEALTH):
            return group.pupil_set.filter(grup=group).count()

        return group.pupil_set.filter(
            Q(children__health_need_id__isnull=False) &
            ~Q(children__health_need__code=HNE.NOT)).count()

    @classmethod
    def _get_group_free_space(cls, group, use_fact_norm):
        """Возвращает количество свободных мест в группе по формуле:
         нормативная наполн./макс. наполн - фактическая наполн. - направления
         Возращает 0, если значение отрицательное.

        :param use_fact_norm: использовать макс. наполняемость
        """
        if use_fact_norm:
            max_count = group.max_count
        else:
            max_count = cls._get_capacity(group)
        fact_count = group.pupil_set.filter(
            grup=group, temporary_deduct=False).count()

        direct_count = Direct.objects.filter(
            Q(group=group) &
            Q(status__code=DRS.REGISTER) |
            Q(status__code=DRS.DOGOVOR) |
            Q(status__code=DRS.NEW) |
            Q(status__code=DRS.CONFIRM)
        ).count()
        result = max_count - fact_count - direct_count

        return result if result > 0 else 0

    @classmethod
    def get_info(cls, dou):
        use_fact_norm = dou.get_use_fact_norm_cnt()
        dou_groups = Group.objects.select_related(
            'type', 'health_need'
        ).filter(unit=dou, status=GroupStatusEnum.FACT)

        result = []
        for group in dou_groups:

            age_sub_category = group.sub_age_cat
            if age_sub_category:
                code = age_sub_category.code
                age_from, age_to = [int(age) for age in code.split('-')]
                if not age_from:
                    age_from += 0.1
            else:
                age_from, age_to = 1, 7

            group_dict = {
                'id': group.id,
                'ageFrom': age_from,
                'ageTo': age_to,
                'ovzType': '',
                'wellness': '',
                'capacity': cls._get_capacity(group),
                'enrolled': cls._get_enrolled(group),
                'work_time_group': cls._get_group_work_time(group),
                'activity': cls._get_group_acitivity(group),
                'free_space': cls._get_group_free_space(group, use_fact_norm),
                'ovz_deti': cls._get_count_ovz_child(group),
            }

            group_dict.update(
                cls._get_group_orientation(group, age_from, age_to))

            if group.name:
                group_dict.update(
                    {'name': group.name}
                )
            else:
                group_dict.update(
                    {"name": u"нет"}
                )

            result.append(group_dict)

        return result


class TemplateLoader(object):
    TEMPLATE = 'ws/gisdo_request_template.xml'

    def __init__(self, schedule_settings):

        self._settings = schedule_settings

    def load(self, document):
        xml_string = loader.render_to_string(
            self.TEMPLATE, document).encode('utf-8')
        if self._settings.zip_xml:
            xml_string = zlib.compress(xml_string)[2:-4]

        encoded_xml = base64.b64encode(xml_string)

        return encoded_xml


class BroadCaster(object):
    def __init__(self, schedule_settings):

        self._settings = schedule_settings
        self._client = self._create_client()

    def _create_client(self):

        try:
            header = Element('Authentication', ns=('ns1', 'http://eo.edu.ru'))
            header.append(
                Element('Login', ns=('ns1', 'http://eo.edu.ru')).setText(
                    self._settings.push_login))
            header.append(
                Element('Password', ns=('ns1', 'http://eo.edu.ru')).setText(
                    self._settings.push_password))
            client = Client(
                url=gs_settings.GISDO_WS_WSDL,
                timeout=gs_settings.GISDO_TIMEOUT)
            client.set_options(soapheaders=(header,))
        except Exception:
            raise SOAPClientCreateException()
        return client

    def _get_send_method(self):

        if self._settings.zip_xml:
            if self._settings.async_send:
                method = 'PushZipDataAsync'
            else:
                method = 'PushZipData'
        else:
            method = 'PushData'

        return method

    def _get_number_of_attempts(self):

        return int(self._settings.resend_count)

    def send(self, content):

        method = self._get_send_method()

        attempts = 1
        error_msg = None
        error_inst = None
        while attempts < self._get_number_of_attempts():
            try:
                result = getattr(self._client.service, method)(
                    data=content, schema=u'3.0')
            except (socket.timeout, suds.WebFault) as error:
                error_inst = error
                error_msg = str(error).encode('utf-8')
                attempts += 1
            except Exception, error:
                error_inst = error
                error_msg = str(error).encode('utf-8')
                break
            else:
                if not result['Result']:
                    error_msg = result['Message']
                break

        if error_msg:
            Logger.add_record(error_msg, level=Logger.ERROR)
            if isinstance(error_inst, (socket.timeout, suds.WebFault)):
                raise ConnectingAtSendingReportException()
            raise SendReportException()


def get_age_filter(age, on_date):

    down, up = AgeDeltas.get_category_deltas(age, on_date)

    filter_ = lambda rec: rec["children__date_of_birth"] > down
    if up:
        filter_ = lambda rec: (
            rec["children__date_of_birth"] > down and
            rec["children__date_of_birth"] <= up
        )

    return filter_


def get_enrolled_age_filter(age, on_date):
    down, up = AgeDeltas.get_category_deltas(age, on_date)

    filter_ = lambda rec: rec > down
    if up:
        filter_ = lambda rec: (down < rec <= up)

    return filter_